/* FLENG compiler - FFI wrapper generation
*/

generate_stubs([], Tail, Tail).
generate_stubs([Spec|Specs], Code, Tail) :-
    atom(Spec),
    assertz(foreign_wrapper('$verbatim', Spec)),
    generate_stubs(Specs, Code, Tail).
generate_stubs([struct(Struct)|Specs], Code, Tail) :-
    generate_struct(Struct, Code, Tail1),
    generate_stubs(Specs, Tail1, Tail).
generate_stubs([const(Const)|Specs], Code, Tail) :-
    generate_const(Const, Code, Tail1),
    generate_stubs(Specs, Tail1, Tail).
generate_stubs([Name = Spec|Specs], Code, Tail) :-
    atom(Name),
    generate_stub(Name, Spec, Code, Tail1),
    generate_stubs(Specs, Tail1, Tail).
generate_stubs([Spec|Specs], Code, Tail) :-
    callable(Spec),
    functor(Spec, Name, _),
    generate_stub(Name, Spec, Code, Tail1),
    generate_stubs(Specs, Tail1, Tail).
generate_stubs([Spec|_], _, _) :-
    error('invalid foreign specification: ~w\n', [Spec]).

% Stub function parsing and generation

generate_stub(Name, Spec, Code, Tail) :-
    Spec =.. [Fname|Args],
    assertz(foreign_wrapper(Fname, Args)),
    parse_foreign_args(Args, 0, Vars, CallArgs, Params, Derefs, no),
    foreign_stub_name(Fname, Stub),
    gen_foreign_stub(Name, Vars, CallArgs, Params, Derefs, Stub, Code, Tail).

foreign_stub_name(Name, SName) :-
    format_to_atom(SName, 'fl_f_~a', [Name]).

parse_foreign_args([], V, [V], ['$VAR'(V)], ['$VAR'(V)], [], no).
parse_foreign_args([], _, [], [], [], [], _).
parse_foreign_args([-void], V, [V], [Var], [Var], [], _) :-
    Var = '$VAR'(V).
parse_foreign_args([-_|Args], V, [V|Vars], [Var|CallArgs], [Var|Params], Derefs, _) :-
    Var = '$VAR'(V),
    V2 is V + 1,
    parse_foreign_args(Args, V2, Vars, CallArgs, Params, Derefs, yes).
parse_foreign_args([T|Args], V, [V|Vars], [Var|CallArgs], ['!'(Var)|Params], 
    [Var|Derefs], R) :-
    memberchk(T, [string, list, tuple]),
    !,
    Var = '$VAR'(V),
    V2 is V + 1,
    parse_foreign_args(Args, V2, Vars, CallArgs, Params, Derefs, R).
parse_foreign_args([T|Args], V, [V|Vars], [Var|CallArgs], ['!'(Var)|Params], Derefs, 
    R) :-
    (memberchk(T, [integer, char, long, short, real, pointer]); T = pointer(_)),
    !,
    Var = '$VAR'(V),
    V2 is V + 1,
    parse_foreign_args(Args, V2, Vars, CallArgs, Params, Derefs, R).
parse_foreign_args([_|Args], V, [V|Vars], [Var|CallArgs], [Var|Params], Derefs, R) :-
    Var = '$VAR'(V),
    V2 is V + 1,
    parse_foreign_args(Args, V2, Vars, CallArgs, Params, Derefs, R).

gen_foreign_stub(Name, Vars, CallArgs, Params, Derefs, Stub, Code, Tail) :-
    Derefs \= [],
    length(Vars, CV),
    format_to_atom(Name2, '$d$~a', [Name]),
    Call2 =.. [Name2, '$VAR'(CV)|CallArgs],
    Head2 =.. [Name2, '!'('$VAR'(CV))|Params],
    canonical_call_args(CallArgs, CA2),
    Call =.. [Stub|CA2],
    Head =.. [Name|Params],
    length(CallArgs, N),
    N2 is N + 1,
    assertz(synthesized_name(Name2/N2, Name/N)),
    Code = [t((Head :- call(lib:deref(Derefs, '$VAR'(CV))), Call2), [], []),
            t((Head2 :- foreign_call(Call)), [], [])|Tail].
gen_foreign_stub(Name, _, CallArgs, Params, [], Stub, Code, Tail) :-
    canonical_call_args(CallArgs, CA2),
    Call =.. [Stub|CA2],
    Head =.. [Name|Params],
    Code = [t((Head :- foreign_call(Call)), [], [])|Tail].

canonical_call_args(Args, CA) :- canonical_call_args(Args, 0, CA).

canonical_call_args([], _, []).
canonical_call_args([X], 2, [X]).
canonical_call_args([X|Args], 2, [X, Args]).
canonical_call_args([X|Args], N, [X|CA]) :-
    N < 2,
    N2 is N + 1,
    canonical_call_args(Args, N2, CA).

% Struct wrapper parsing

generate_struct(Name = Spec, Code, Tail) :-
    atom(Name),
    generate_struct2(Name, Spec, Code, Tail).
generate_struct(Spec, Code, Tail) :-
    callable(Spec),
    functor(Spec, Name, _),
    generate_struct2(Name, Spec, Code, Tail).
generate_struct(Spec, _, _) :-
    error('invalid foreign struct specification: ~w\n', [Spec]).

generate_struct2(Name, Spec, C, T) :-
    Spec =.. [Sname|Args],
    generate_struct_accessors(Args, Name, Sname, C, T).

generate_struct_accessors([], _, _, C, C).
generate_struct_accessors([-Slot:Type|Args], Name, Sname, C, T) :-
    atom(Slot),
    generate_struct_getter(Name, Sname, Slot, Type, C, T1),
    generate_struct_accessors(Args, Name, Sname, T1, T).
generate_struct_accessors([Slot:Type|Args], Name, Sname, C, T) :-
    atom(Slot),
    generate_struct_getter(Name, Sname, Slot, Type, C, T1),
    generate_struct_setter(Name, Sname, Slot, Type, T1, T2),
    generate_struct_accessors(Args, Name, Sname, T2, T).
generate_struct_accessors([Sspec|_], _, Sname, _, _) :-
    error('invalid slot specification for foreign struct ~a: ~w\n', [Sspec, Sname]).

generate_struct_getter(Name, Sname, Slot, Type, C, Tail) :-
    format_to_atom(GetN, 's_g_~a_~a', [Name, Slot]),
    format_to_atom(Get, '#define ~a(_x) (_x->~a)', [GetN, Slot]),
    assertz(foreign_wrapper('$verbatim', Get)),
    format_to_atom(Getter, '~a_~a', [Name, Slot]),
    Spec =.. [GetN, pointer(Sname), -Type],
    generate_stub(Getter, Spec, C, Tail).

generate_struct_setter(Name, Sname, Slot, Type, C, Tail) :-
    format_to_atom(SetN, 's_s_~a_~a', [Name, Slot]),
    format_to_atom(Set, '#define ~a(_x, _y) (_x->~a = _y)', [SetN, Slot]),
    assertz(foreign_wrapper('$verbatim', Set)),
    format_to_atom(Setter, '~a_~a', [Name, Slot]),
    Spec =.. [SetN, pointer(Sname), Type],
    generate_stub(Setter, Spec, C, Tail).

% Constant stubs

generate_const(Name = Const:Type, C, T) :-
    generate_const2(Name, Const, Type, C, T).
generate_const(Name:Type, C, T) :-
    generate_const2(Name, Name, Type, C, T).
generate_const(X, _, _) :-
    error('invalid foreign const specification: ~w\n', [X]).

generate_const2(Name, Const, Type, C, Tail) :-
    format_to_atom(GetC, 'c_g_~a', [Name]),
    native_type(Type, NT),
    format_to_atom(Get, 'static ~a ~a() {return ~a;}', [NT, GetC, Const]),
    assertz(foreign_wrapper('$verbatim', Get)),
    Spec =.. [GetC, -Type],
    generate_stub(Name, Spec, C, Tail).

% C wrapper code generation

generate_wrappers :-
    foreign_wrapper(_, _),
    !,
    wrapper_file_name(Fname),
    generate_wrappers(Fname).
generate_wrappers.

wrapper_file_name(Fname) :-
    module_name(Mod),
    format_to_atom(Fname, '~a.c', [Mod]).
wrapper_file_name('foreign.c').

generate_wrappers(Fname) :-
    open(Fname, write, Out),
    format(Out, '/* GENERATED BY FLENG */\n#include "fleng-util.h"\n#include <string.h>\n', []),
    write_wrappers(Out),
    format(Out, '/* END OF FILE */\n', []),
    close(Out).

write_wrappers(Out) :-
    foreign_wrapper(Name, Args),
    write_wrapper(Name, Args, Out),
    fail.
write_wrappers(_).

write_wrapper('$verbatim', Text, Out) :-
    !,
    format(Out, '~a\n', [Text]).
write_wrapper(Name, Args, Out) :-
    wrapper_arity(Args, A1, Confirm),
    A is min(A1, 4),
    format(Out, 'void fl_f_~a_~d(FL_TCB *tcb', [Name, A]),
    write_wrapper_params(Args, 0, no, Out),
    format(Out, ') {\n', []),
    write_wrapper_init(Args, Out),
    write_wrapper_conv(Args, 0, A1, Out),
    write_wrapper_call(Name, Args, Out),
    write_wrapper_confirm(Confirm, A1, Out),
    format(Out, '}\n', []),
    !.

write_wrapper_init(Args, Out) :-
    (memberchk(charlist, Args); memberchk(string, Args)),
    !,
    format(Out, '  int sl;\n  char *sp = init_cbuf(tcb);\n', []).
write_wrapper_init(_, _).

wrapper_arity(Args, A, C) :- wrapper_arity(Args, no, 0, A, C).

wrapper_arity([], yes, A, A, no).
wrapper_arity([], no, A1, A, yes) :- A is A1 + 1.
wrapper_arity([-void], no, A1, A, yes) :- A is A1 + 1.
wrapper_arity([-void], yes, A, A, no).
wrapper_arity([-_|Args], _, A1, A, C) :- 
    A2 is A1 + 1,
    wrapper_arity(Args, yes, A2, A, C).
wrapper_arity([_|Args], R, A1, A, C) :-
    A2 is A1 + 1,
    wrapper_arity(Args, R, A2, A, C).

write_wrapper_confirm(yes, A, Out) :-
    N is A - 1,
    format(Out, '  fl_assign(tcb, fl_nil, p~d);\n', [N]).
write_wrapper_confirm(no, _, _).

write_wrapper_params([], N, no, Out) :-
    N < 3,
    format(Out, ', FL_VAL p~d', [N]).    
write_wrapper_params([], _, _, _).
write_wrapper_params([T], 3, no, Out) :-
    (T = -void; T \= -_),
    format(Out, ', FL_VAL p3', []).
write_wrapper_params([_, _|_], 3, _, Out) :-
    format(Out, ', FL_VAL p3', []).
write_wrapper_params([-_|Args], N, _, Out) :-
    N2 is N + 1,
    format(Out, ', FL_VAL p~d', [N]),
    write_wrapper_params(Args, N2, yes, Out).
write_wrapper_params([_|Args], N, R, Out) :-
    N2 is N + 1,
    format(Out, ', FL_VAL p~d', [N]),
    write_wrapper_params(Args, N2, R, Out).

write_wrapper_conv([], _, _, _).
write_wrapper_conv([-_], N, A, Out) :-
    wrapper_accessor(N, A, _, Out).
write_wrapper_conv([T|Args], N, A, Out) :-
    memberchk(T, [integer, long, char, short]),
    wrapper_accessor(N, A, Var, Out),
    native_type(T, NT),
    format(Out, '  CHECK_INT(~a);\n  ~a a~d = (~a)INT(~a);\n', [Var, NT, N, NT, Var]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([string|Args], N, A, Out) :-
    wrapper_accessor(N, A, Var, Out),
    format(Out, '  CHECK_STRING(~a);\n  char *a~d = stringify_next(tcb, &sp, ~a, &sl);\n  sp += sl + 1;\n', 
        [Var, N, Var]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([charlist|Args], N, A, Out) :-
    wrapper_accessor(N, A, Var, Out),
    format(Out, '  char *a~d = stringify_next(tcb, &sp, ~a, &sl);\n  sp += sl + 1;\n', 
        [N, Var]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([real|Args], N, A, Out) :-
    wrapper_accessor(N, A, Var, Out),
    format(Out, '  CHECK_NUMBER(~a);\n  FLOAT(~a, a~d);\n', [Var, Var, N]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([pointer|Args], N, A, Out) :-
    wrapper_accessor(N, A, Var, Out),
    format(Out, '  CHECK_INT(~a);\n  void *a~d = (void *)POINTER(~a);\n', [Var, N, Var]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([pointer(T)|Args], N, A, Out) :-
    wrapper_accessor(N, A, Var, Out),
    format(Out, '  CHECK_INT(~a);\n  ~a *a~d = (~a *)POINTER(~a);\n', [Var, T, N, T, Var]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([-T|Args], N, A, Out) :-
    memberchk(T, [string, charlist]),
    wrapper_accessor(N, A, _, Out),
    format(Out, '  char *a~d;\n', [N]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([-T|Args], N, A, Out) :-
    (memberchk(T, [integer, long, char, short, pointer, real]); T = pointer(_)),
    !,
    wrapper_accessor(N, A, _, Out),
    native_type(T, NT),
    format(Out, '  ~a a~d;\n', [NT, N]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([tuple|Args], N, A, Out) :-
    wrapper_accessor(N, A, Var, Out),
    format(Out, '  CHECK_TUPLE(~a);\n  FL_VAL a~d = ~a;\n', [Var, N, Var]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([T|Args], N, A, Out) :-
    memberchk(T, [list, tuple, any]),
    wrapper_accessor(N, A, Var, Out),
    format(Out, '  FL_VAL a~d = ~a;\n', [N, Var]),
    N2 is N + 1,
    write_wrapper_conv(Args, N2, A, Out).
write_wrapper_conv([T|_], _, _, _) :-
    error('invalid foreign type: ~w\n', [T]).

wrapper_accessor(N, _, Var, _) :-
    N < 3,
    format_to_atom(Var, 'p~d', [N]).
wrapper_accessor(3, A, Var, Out) :-
    A > 3,
    format(Out, '  FL_VAL pr = CDR(p3);\n  p3 = CAR(p3);\n', []),
    format_to_atom(Var, 'p3', []).
wrapper_accessor(N, A, Var, Out) :-
    N > 3, A > 3,
    format(Out, '  LIST(pr, p~d, r~d);\n  pr = r~d;\n', [N, N, N]),
    format_to_atom(Var, 'p~d', [N]).
wrapper_accessor(_, _, p3, _).

write_wrapper_call(Name, Args, Out) :-
    has_result(Args, Type),
    !,
    length(Args, N),
    N2 is N - 1,
    format(Out, '  ~a a~d = ~a(', [Type, N2, Name]),
    write_wrapper_args(Args, 0, Out),
    format(Out, ');\n', []),
    write_wrapper_results(Args, 0, Out).
write_wrapper_call(Name, Args, Out) :-
    format(Out, '  ~a(', [Name]),
    write_wrapper_args(Args, 0, Out),
    format(Out, ');\n', []),
    write_wrapper_results(Args, 0, Out).

has_result([-void], _) :- !, fail.
has_result([-integer], long) :- !.
has_result([-T], NT) :- native_type(T, NT), !.
has_result([_|Args], T) :- has_result(Args, T).

native_type(integer, int).
native_type(long, long).
native_type(short, short).
native_type(string, 'char *').
native_type(pointer, 'void *').
native_type(real, double).
native_type(char, char).
native_type(pointer(T), NT) :- format_to_atom(NT, '~a *', [T]).
native_type(_, 'FL_VAL').

write_wrapper_args([], _, _).
write_wrapper_args([-_], _, _).
write_wrapper_args([-_, T|Args], N, Out) :-
    (T \= -_; Args \= []),
    !,
    format(Out, '&a~d, ', [N]),
    N2 is N + 1,
    write_wrapper_args(Args, N2, Out).
write_wrapper_args([-_|Args], N, Out) :-
    format(Out, '&a~d', [N]),
    N2 is N + 1,
    write_wrapper_args(Args, N2, Out).
write_wrapper_args([_], N, Out) :-
    format(Out, 'a~d', [N]).
write_wrapper_args([_, -_], N, Out) :-
    format(Out, 'a~d', [N]).
write_wrapper_args([_|Args], N, Out) :-
    format(Out, 'a~d, ', [N]),
    N2 is N + 1,
    write_wrapper_args(Args, N2, Out).

write_wrapper_results([], _, _).
write_wrapper_results([-void], _, _).
write_wrapper_results([-T|Args], N, Out) :-
    memberchk(T, [char, integer, long, short]),
    !,
    format(Out, '  fl_unify_result(tcb, MKINT(a~d), p~d);\n', [N, N]),
    N2 is N + 1,
    write_wrapper_results(Args, N2, Out).
write_wrapper_results([-real|Args], N, Out) :-
    format(Out, '  fl_unify_result(tcb, fl_alloc_float(tcb, a~d), p~d);\n', [N, N]),
    N2 is N + 1,
    write_wrapper_results(Args, N2, Out).
write_wrapper_results([-string|Args], N, Out) :-
    format(Out, '  fl_unify_result(tcb, fl_mkstring(a~d, strlen(a~d)), p~d);\n', 
        [N, N, N]),
    N2 is N + 1,
    write_wrapper_results(Args, N2, Out).
write_wrapper_results([-T|Args], N, Out) :-
    (T = pointer; T = pointer(_)),
    !,
    format(Out, '  fl_unify_result(tcb, MKINT((unsigned long)a~d), p~d);\n', 
        [N, N]),
    N2 is N + 1,
    write_wrapper_results(Args, N2, Out).
write_wrapper_results([-charlist|Args], N, Out) :-
    format(Out, '  fl_unify_result(tcb, mkcharlist(tcb, a~d, strlen(a~d), fl_nil), p~d);\n', 
        [N, N, N]),
    N2 is N + 1,
    write_wrapper_results(Args, N2, Out).
write_wrapper_results([-_|Args], N, Out) :-
    !,
    format(Out, '  fl_unify_result(tcb, a~d, p~d);\n', [N, N]),
    N2 is N + 1,
    write_wrapper_results(Args, N2, Out).
write_wrapper_results([_|Args], N, Out) :-
    N2 is N + 1,
    write_wrapper_results(Args, N2, Out).
